from .styling import sep

header_error = "\n\n-- ! ERROR ! --\n"
header_warning = "\n-- ! WARNING ! --\n"
header_gmx_error = "\n\n  GMX ERROR MSG\n" + sep + "\n\n"

# TODO: disable showing exceptions traceback for distributing to users ? maybe there is a way we can keep it active for our dev purpose though ?

class BaseError(Exception):
    """
    Base exception class.
    """
    def __init__(self, msg):
        self.message = header_error + msg
        super().__init__(self.message)


class ExecError(Exception):
    """
    Base exception class.
    """
    def __init__(self, msg):
        self.message = header_gmx_error + msg
        super().__init__(self.message)


class InvalidArgument(BaseError):

    base_msg = (
        "Cannot interpret argument -{name} as provided: {value}. "
        "{additional_info}"
        "Please check your parameters, or look for help in an example."
    )

    def __init__(self, argument, value, additional_info=""):
        self.argument = argument
        self.value = value
        self.additional_info = additional_info
        self.message = self.base_msg.format(
            name=argument, value=value, additional_info=additional_info
        )
        super().__init__(self.message)


class IncompleteOptimisationFile(BaseError):
    pass


class OptimisationResultsError(BaseError):
    pass


class MissingCoordinateFile(BaseError):
    pass


class MissingTrajectoryFile(BaseError):
    pass


class MissingItpFile(BaseError):
    pass


class MissingIndexFile(BaseError):
    pass


class InputArgumentError(BaseError):
    pass


class ExecutableNotFound(BaseError):
    pass


class AvoidOverwritingFolder(BaseError):
    pass


class MDSimulationInputError(BaseError):
    pass


class MissformattedFile(BaseError):
    pass


class ComputationError(ExecError):
    pass


import numpy as np


def gmx_bonds_func_1(x, a, b, c):
	"""
	Gromacs potential function 1 for bonds.
	"""
	return a / 2 * (x - b) ** 2 + c


def gmx_angles_func_1(x, a, b, c):
	"""
	Gromacs potential function 1 for angles.
	"""
	return gmx_bonds_func_1(x, a, b, c)


def gmx_angles_func_2(x, a, b, c):
	"""
	Gromacs potential function 2 for angles.
	"""
	return a / 2 * (np.cos(x) - np.cos(b)) ** 2 + c


def gmx_dihedrals_func_1(mult):
	"""
	Gromacs potential function 1 for angles -- generated on the fly with adjusted multiplicity
	"""
	def mult_adjusted(x, a, b, c):
		return a * (1 + np.cos(mult * x - b)) + c
	return mult_adjusted


def gmx_dihedrals_func_2(x, a, b, c):
	"""
	Gromacs potential function 2 for dihedrals -- the same as potential function 1 for angles
	"""
	return gmx_bonds_func_1(x, a, b, c) # it's actually the same


# TODO: for dihedral function 9, this is the merging of several potentials of
# TODO: gmx_dihedrals_func_1 -- here one of mult=1 together with another of mult=2
# def f(x,a,b,c,d,e):
# 	return a * (1+np.cos(x-b)) + d * (1+np.cos(2*x-e)) + c


# general stuff
github_url = 'http://github.com/GMPavanLab/SwarmCG'
gmx_path = 'gmx'

# BI and FST-PSO OPTI, defaults
kB = 0.008314462
sim_temperature = 300  # Kelvin
bi_nb_bins = 50  # nb of bins to use for Boltzmann Inversion, will be doubled for dihedrals distributions binning during BI -- this has huge impact on the results of the BI and this value shall STAY AT 50 ! actually I did not try to modify much but this feels like dangerous atm
bonds_max_range = 15  # nm -- used to define grid for EMD calculations
# NOTE: increasing bonds_max_range increases computation time, but memory usage increases exponentially
# TODO: detect when a bond has longer values than bonds_max_range and suggest the user to raise the limit if he really
#       needs to, but I don't really see what kind of use case would require more than 5 nm (maybe elastic net in
#       proteins though)
bw_constraints = 0.002  # nm
bw_bonds = 0.01  # nm
bw_angles = 2.5  # degrees
bw_dihedrals = 2.5  # degrees
default_min_fct_bonds = 0
default_max_fct_bonds_bi = 17000
default_max_fct_bonds_opti = 18000
default_min_fct_angles = 0
default_max_fct_angles_bi = 1200
default_max_fct_angles_opti_f1 = 1700
default_max_fct_angles_opti_f2 = 1700

default_min_fct_dihedrals_func_without_mult = 0
default_max_fct_dihedrals_bi_func_without_mult = 250
default_max_fct_dihedrals_opti_func_without_mult = 1500

# default_max_fct_dihedrals_func_with_mult = 0
# default_min_fct_dihedrals_bi_func_with_mult = -10
# default_min_fct_dihedrals_opti_func_with_mult = -15 # for dihedrals it makes sense to have negative force constants for functions that use multiplicity

default_abs_range_fct_dihedrals_bi_func_with_mult = 3.5
default_abs_range_fct_dihedrals_opti_func_with_mult = 15

bonds2angles_scoring_factor = 250  # multiplier applied to constraints/bonds EMD scores to retrieve angles/dihedrals mismatches that are comparable, for the opti scoring function
sim_crash_EMD_indep_score = 150  # when a simulation crashes or does not finish for any reason: EMD distance between 2 distributions, for 1 geom

# bonds scaling, default
bonds_scaling = 1.0  # ratio
min_bonds_length = 0.00  # nm
bonds_scaling_str = ''  # constraints and bonds ids + their required target AA-mapped distributions rescaled averages

# NOTE: in the benchmark of the paper we used bonds2_angles_scoring_factor = 500 and fct_guess_min_flat_diff_angles = 50,
#       differences in the results should be small but still an improvement

# building of the initial guesses for optimization, defaults
bond_dist_guess_variation = 0.025  # nm
angle_value_guess_variation = 10  # degrees
dihedral_value_guess_variation = 10  # degrees
fct_guess_min_flat_diff_bonds = 200  # flat minimum force constant variation that fct_guess_fact shall yield, used to find low and high boundaries for random generation of particles' force constants
fct_guess_min_flat_diff_angles = 100  # flat minimum force constant variation that fct_guess_fact shall yield, used to find low and high boundaries for random generation of particles' force constants
fct_guess_min_flat_diff_dihedrals_without_mult = 0.50  # flat minimum force constant variation that fct_guess_fact shall yield, used to find low and high boundaries for random generation of particles' force constants
fct_guess_min_flat_diff_dihedrals_with_mult = 0.20  # flat minimum force constant variation that fct_guess_fact shall yield, used to find low and high boundaries for random generation of particles' force constants

# gromacs functions that are properly treated at the moment
# if we find a function that is not handled, program will exit with an appropriate error message
# TODO: handle dihedral function 9 correctly so that different potentials can be stacked for the same beads
#       this is the primary purpose of function 9 !!
handled_functions = {
    'constraint': [1],  # tested and verified: 1
    'bond': [1],  # tested and verified: 1
    'angle': [1, 2],  # tested and verified: 1, 2
    'dihedral': [1, 2, 4],  # tested and verified: 1, 2, 4 -- ongoing: 9 (need to merge the 1+ dihedrals groups on plots)
    'virtual_sites2': [1],  # tested and verified: 1 -- ongoing: 2 (need GMX 2020)
    'virtual_sites3': [1, 2, 3, 4],  # tested and verified: 1, 2, 3, 4
    'virtual_sites4': [2],  # tested and verified: 2 -- irrelevant: 1
    'virtual_sitesn': [1, 2, 3]  # tested and verified: 1, 2, 3
}
dihedral_func_with_mult = [1, 4, 9]  # these functions use 3 parameters, the last one being multiplicity

# plots display parameters
use_hists = False  # hists are not implemented in a way that they will be displayed with left and right bold borders atm
line_alpha = 0.6  # line alpha for the density plots
fill_alpha = 0.30  # fill alpha for the density plots
cg_color = '#1f77b4'
atom_color = '#d62728'

# scripts inputs default filenames
metavar_aa_tpr = 'aa_topol.tpr'
metavar_aa_traj = 'aa_traj.xtc'
metavar_cg_map = 'cg_map.ndx'
metavar_cg_itp = 'cg_model.itp'
metavar_cg_tpr = 'cg_topol.tpr'
metavar_cg_traj = 'cg_traj.xtc'

# help descriptions for arguments
help_aa_tpr = 'Topology binary file of your reference AA simulation (TPR)'
help_aa_traj = 'Trajectory file of the reference AA simulation (XTC, TRR)\nPBC are handled internally if trajectory contains box dimensions'
help_cg_map = 'Mapping file of the atoms to CG beads (NDX-like file format)'
help_mapping_type = 'Center Of Mass (COM) or Center Of Geometry (COG), for\ninterpreting the mapping file'
help_verbose = 'Display more processing details & error traceback'
help_gmx_path = 'Your Gromacs alias/path'
help_bonds_scaling = 'Scaling factor for ALL AA-mapped bonds/constraints lengths\nOnly one of arguments -bonds_scaling, -bonds_scaling_str\nand -min_bonds_length can be provided'
help_min_bonds_length = 'Required minimum length of a bond or constraint between 2 CG\nbeads (distributions avg in nm), used both as:\n1. Threshold to identify ALL short AA-mapped bonds/constraints\n2. Target avg to rescale ALL those bonds/constraints'
help_bonds_scaling_str = 'String (use quotes) for providing SPECIFIC bonds/constraints\ngroups ids and their required lengths (nm, rescaled\ndistributions avg to use as target for optimization)\nEx: \'C1 0.23 B5 0.27\' will modify distributions of constraints\ngrp 1 and bonds grp 5 to averages 0.23 and 0.27 nm'
help_bonds2angles_scoring_factor = 'Weight of bonds vs. angles/dihedrals (constant C in the paper)'
help_bw_constraints = 'Bandwidth for constraints distributions processing (nm)'
help_bw_bonds = 'Bandwidth for bonds distributions processing (nm)'
help_bw_angles = 'Bandwidth for angles distributions processing (degrees)'
help_bw_dihedrals = 'Bandwidth for dihedrals distributions processing (degrees)'
help_bonds_max_range = 'Max. range of grid for bonds/constraints distributions (nm)'
help_max_fct_bonds = 'Max. force constants for bonds function 1 (kJ mol-1 nm-2)'
help_max_fct_angles_f1 = 'Max. force ct. for angles function 1 (kJ mol-1 rad-2)'
help_max_fct_angles_f2 = 'Max. force ct. for angles function 2 (kJ mol-1)'
help_max_fct_dihedrals_with_mult = 'Max. force ct. for dihedrals functions 1, 4, 9 (abs. kJ mol-1)'
help_max_fct_dihedrals_without_mult = 'Max. force ct. for dihedrals function 2 (kJ mol-1 rad-2)'

# optimization output filenames
input_sim_files_dirname = '.internal/input_CG_simulation_files'
iteration_sim_files_dirname = 'CG_sim_files'  # basename to be appended to with _NN
best_fitted_model_dirname = 'optimized_CG_model'
distrib_plots_all_evals_dirname = 'all_evals_distributions'
log_files_all_evals_dirname = 'all_evals_logs'
sim_files_all_evals_dirname = 'all_evals_all_sim_files'
opti_perf_recap_file = '.internal/opti_recap_evals_perfs_and_params.csv'
opti_pairwise_distances_file = '.internal/opti_recap_evals_pairwise_distribs_diffs.csv'
ref_distrib_plots = 'reference_AA_distributions.png'
best_distrib_plots = 'optimized_CG_model_distributions.png'










from setuptools import setup, find_packages

import swarmcg

with open('README.md') as readme_file:
    README = readme_file.read()

setup(
    name='swarm-cg',
    version=swarmcg.__version__,
    description='Tools for automatic parametrization of bonded terms in coarse-grained molecular models, with respect to an all-atom trajectory',
    author='Charly Empereur-mot',
    author_email='charly.empereur@gmail.com',
    license='MIT',
    url='https://github.com/GMPavanLab/Swarm-CG',
    download_url=f'https://github.com/GMPavanLab/Swarm-CG/archive/v{swarmcg.__version__}.tar.gz',
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=['gromacs', 'coarse-grain', 'molecular model', 'optimization', 'force field'],
    packages=find_packages(include=['swarmcg', 'swarmcg.*']),
    install_requires=[
        'numpy>=1.16.4',
        'scipy>=1.2.2',
        'pyemd>=0.5.1',
        'matplotlib>=3.2.2',  # warnings appear for lower versions, don't change
        'fst-pso>=1.4.12',
        'MDAnalysis>=1.0.0'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    entry_points={
        'console_scripts': [
            'scg_optimize=swarmcg.optimize_model:main',
            'scg_evaluate=swarmcg.evaluate_model:main',
            'scg_monitor=swarmcg.analyze_optimization:main'
        ]
    }
)






import numpy as np
import MDAnalysis as mda
from ..shared import exceptions

# All these functions for virtual sites definitions are explained
# in the GROMACS manual part 5.5.7 (page 379 in manual version 2020)
# Check also the bonded potentials table best viewed here:
# http://manual.gromacs.org/documentation/2020/reference-manual/topologies/topology-file-formats.html#tab-topfile2

# TODO: test all these functions


# Functions for virtual_sites2

# vs_2 func 1 -> Linear combination using 2 reference points
# weighted COG using a percentage in [0, 1]
# the weight is applied on the bead ID that comes first
def vs2_func_1(ns, traj, vs_def_beads_ids, vs_params):

    i, j = vs_def_beads_ids
    a = vs_params  # weight
    weights = np.array([1-a, a])

    for ts in ns.aa2cg_universe.trajectory:
        traj[ts.frame] = ns.aa2cg_universe.atoms[[i, j]].center(weights)


# vs_2 func 2 -> Linear combination using 2 reference points
# on the vector from i to j, at given distance (nm)
# NOTE: it seems this one exists only since GROMACS 2020
# TODO: check this one with a GMX 2020 installation
def vs2_func_2(ns, traj, vs_def_beads_ids, vs_params):

    i, j = vs_def_beads_ids
    a = vs_params  # nm
    a = a * 10  # retrieve amgstrom for MDA

    for ts in ns.aa2cg_universe.trajectory:
        pos_i = ns.aa2cg_universe.atoms[i].position
        pos_j = ns.aa2cg_universe.atoms[j].position
        r_ij = pos_j - pos_i
        traj[ts.frame] = pos_i + a * r_ij / mda.lib.mdamath.norm(r_ij)


# Functions for virtual_sites3

# vs_3 func 1 -> Linear combination using 3 reference points
# in the plane, using sum of vectors from i to j and from k to i
def vs3_func_1(ns, traj, vs_def_beads_ids, vs_params):

    i, j, k = vs_def_beads_ids
    a, b = vs_params  # nm, nm
    a, b = a * 10, b * 10  # retrieve amgstrom for MDA

    for ts in ns.aa2cg_universe.trajectory:
        pos_i = ns.aa2cg_universe.atoms[i].position
        pos_j = ns.aa2cg_universe.atoms[j].position
        pos_k = ns.aa2cg_universe.atoms[k].position
        r_ij = pos_j - pos_i
        r_ik = pos_k - pos_i
        traj[ts.frame] = pos_i + a * r_ij / mda.lib.mdamath.norm(r_ij) / 2 + b * r_ik / mda.lib.mdamath.norm(r_ik) / 2


# vs_3 func 2 -> Linear combination using 3 reference points
# in the plane, using WEIGHTS sum of vectors from j to i and from k to i + fixed distance
# I used their formula (hopefully) so the form differs from the explanation on line above, but it should be identical
def vs3_func_2(ns, traj, vs_def_beads_ids, vs_params):

    i, j, k = vs_def_beads_ids
    a, b = vs_params  # weight, nm
    b = b * 10  # retrieve amgstrom for MDA

    for ts in ns.aa2cg_universe.trajectory:
        pos_i = ns.aa2cg_universe.atoms[i].position
        pos_j = ns.aa2cg_universe.atoms[j].position
        pos_k = ns.aa2cg_universe.atoms[k].position
        r_ij = pos_j - pos_i
        r_jk = pos_k - pos_j
        comb_ijk = (1-a) * r_ij + a * r_jk
        traj[ts.frame] = pos_i + b * (comb_ijk / mda.lib.mdamath.norm(comb_ijk))


# vs_3 func 3 -> Linear combination using 3 reference points
# angle in the plane defined, at given distance of the 3rd point
def vs3_func_3(ns, traj, vs_def_beads_ids, vs_params):

    i, j, k = vs_def_beads_ids
    ang_deg, d = vs_params  # degrees, nm
    ang_rad = np.deg2rad(ang_deg)  # retrieve radians
    d = d * 10  # retrieve amgstrom for MDA

    for ts in ns.aa2cg_universe.trajectory:
        pos_i = ns.aa2cg_universe.atoms[i].position
        pos_j = ns.aa2cg_universe.atoms[j].position
        pos_k = ns.aa2cg_universe.atoms[k].position
        r_ij = pos_j - pos_i
        r_jk = pos_k - pos_j
        comb_ijk = r_jk - (np.dot(r_ij, r_jk) / np.dot(r_ij, r_ij)) * r_ij
        traj[ts.frame] = pos_i + d * np.cos(ang_rad) * (r_ij / mda.lib.mdamath.norm(r_ij)) + d * np.sin(ang_rad) * (comb_ijk / mda.lib.mdamath.norm(comb_ijk))


# vs_3 func 4 -> Linear combination using 3 reference points
# out of plane
def vs3_func_4(ns, traj, vs_def_beads_ids, vs_params):

    i, j, k = vs_def_beads_ids
    a, b, c = vs_params  # weight, weight, nm**(-1)
    c = c / 10  # retrieve amgstrom**(-1) for MDA

    for ts in ns.aa2cg_universe.trajectory:
        pos_i = ns.aa2cg_universe.atoms[i].position
        pos_j = ns.aa2cg_universe.atoms[j].position
        pos_k = ns.aa2cg_universe.atoms[k].position
        r_ij = pos_j - pos_i
        r_ik = pos_k - pos_i
        traj[ts.frame] = pos_i + a * r_ij + b * r_ik - c * (r_ij / mda.lib.mdamath.norm(r_ij) * r_ik / mda.lib.mdamath.norm(r_ik))


# Functions for virtual_sites4

# vs_4 func 2 -> Linear combination using 3 reference points
# NOTE: only function 2 is defined for vs_4 in GROMACS, because it replaces function 1
#       which still exists for retro compatibility but its usage must be avoided
def vs4_func_2(ns, traj, vs_def_beads_ids, vs_params):

    i, j, k, l = vs_def_beads_ids
    a, b, c = vs_params  # weight, weight, nm
    c = c * 10  # retrieve amgstrom for MDA

    for ts in ns.aa2cg_universe.trajectory:
        pos_i = ns.aa2cg_universe.atoms[i].position
        pos_j = ns.aa2cg_universe.atoms[j].position
        pos_k = ns.aa2cg_universe.atoms[k].position
        pos_l = ns.aa2cg_universe.atoms[l].position
        r_ij = pos_j - pos_i
        r_ik = pos_k - pos_i
        r_il = pos_l - pos_i
        r_ja = a * r_ik - r_ij
        r_jb = b * r_il - r_ij
        r_m = np.cross(r_ja, r_jb)
        traj[ts.frame] = pos_i - c * (r_m / mda.lib.mdamath.norm(r_m))


# Functions for virtual_sitesn

# vs_n func 1 -> Center of Geometry
def vsn_func_1(ns, traj, vs_def_beads_ids):

    for ts in ns.aa2cg_universe.trajectory:
        traj[ts.frame] = ns.aa2cg_universe.atoms[vs_def_beads_ids].center_of_geometry(pbc=None)


# vs_n func 2 -> Center of Mass
def vsn_func_2(ns, traj, vs_def_beads_ids, bead_id):

    # inform user if this VS definition uses beads (or VS) with mass 0,
    # because this is COM so 0 mass means a bead that was marked for defining the VS is in fact ignored
    zero_mass_beads_ids = []
    for bid in vs_def_beads_ids:
        if bid in ns.cg_itp['virtual_sitesn']:
            if ns.cg_itp['virtual_sitesn'][bid]['mass'] == 0:
                zero_mass_beads_ids.append(bid)
    if len(zero_mass_beads_ids) > 0:
        print('  WARNING: Virtual site ID {} uses function 2 for COM, but its definition contains IDs ' + ' '.join(zero_mass_beads_ids) + 'which have no mass'.format(bead_id + 1))

    for ts in ns.aa2cg_universe.trajectory:
        traj[ts.frame] = ns.aa2cg_universe.atoms[vs_def_beads_ids].center_of_mass(pbc=None)


# vs_n func 3 -> Center of Weights (each atom has a given weight, pairwise formatting: id1 w1 id2 w2 ..)
def vsn_func_3(ns, traj, vs_def_beads_ids, vs_params):

    masses_and_weights = np.array([ns.aa2cg_universe.atoms[vs_def_beads_ids[i]].mass * vs_params[i] for i in range(len(vs_def_beads_ids))])
    for ts in ns.aa2cg_universe.trajectory:
        traj[ts.frame] = ns.aa2cg_universe.atoms[vs_def_beads_ids].center(masses_and_weights)






